# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from gettext import gettext as _

from sqlalchemy import asc, desc, func
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound

from . import BaseAlchemyManager, query_apply_limit_offset, query_apply_true_or_not
from ....managers.category import BaseCategoryManager
from ..objects import AlchemyActivity, AlchemyCategory, AlchemyFact


class CategoryManager(BaseAlchemyManager, BaseCategoryManager):
    def get_or_create(self, category, raw=False, skip_commit=False):
        """
        Custom version of the default method in order to provide access
        to alchemy instances.

        Args:
            category (nark.Category): Category we want.
            raw (bool): Wether to return the AlchemyCategory instead.

        Returns:
            nark.Category or None: Category.
        """

        message = _("Received {!r} and raw={}.".format(category, raw))
        self.store.logger.debug(message)

        try:
            category = self.get_by_name(category.name, raw=raw)
        except KeyError:
            category = self._add(category, raw=raw, skip_commit=skip_commit)
        return category

    def _add(self, category, raw=False, skip_commit=False):
        """
        Add a new category to the database.

        This method should not be used by any client code. Call ``save`` to
        make the decission wether to modify an existing entry or to add a new
        one is done correctly.

        Args:
            category (nark.Category): nark Category instance.
            raw (bool): Wether to return the AlchemyCategory instead.

        Returns:
            nark.Category: Saved instance, as_hamster()

        Raises:
            ValueError: If the name to be added is already present in the db.
            ValueError: If category passed already got an PK. Indicating that
                update would be more appropriate.
        """
        self.adding_item_must_not_have_pk(category)

        alchemy_category = AlchemyCategory(
            pk=None,
            name=category.name,
            deleted=bool(category.deleted),
            hidden=bool(category.hidden),
        )

        result = self.add_and_commit(
            alchemy_category, raw=raw, skip_commit=skip_commit,
        )

        return result

    def _update(self, category):
        """
        Update a given Category.

        Args:
            category (nark.Category): Category to be updated.

        Returns:
            nark.Category: Updated category.

        Raises:
            ValueError: If the new name is already taken.
            ValueError: If category passed does not have a PK.
            KeyError: If no category with passed PK was found.
        """

        message = _("Received {!r}.".format(category))
        self.store.logger.debug(message)

        if not category.pk:
            message = _(
                "The category passed ('{!r}') does not seem to havea PK. We don't know"
                "which entry to modify.".format(category)
            )
            self.store.logger.error(message)
            raise ValueError(message)
        alchemy_category = self.store.session.query(AlchemyCategory).get(category.pk)
        if not alchemy_category:
            message = _("No category with PK: {} was found!".format(category.pk))
            self.store.logger.error(message)
            raise KeyError(message)
        alchemy_category.name = category.name

        try:
            self.store.session.commit()
        except IntegrityError as e:
            message = _(
                "An error occured! Is category.name already present in the database?"
                " / Error: '{}'.".format(e)
            )
            self.store.logger.error(message)
            raise ValueError(message)

        return alchemy_category.as_hamster(self.store)

    def remove(self, category):
        """
        Delete a given category.

        Args:
            category (nark.Category): Category to be removed.

        Returns:
            None: If everything went alright.

        Raises:
            KeyError: If the ``Category`` can not be found by the backend.
            ValueError: If category passed does not have an pk.
        """

        message = _("Received {!r}.".format(category))
        self.store.logger.debug(message)

        if not category.pk:
            message = _("PK-less Category. Are you trying to remove a new Category?")
            self.store.logger.error(message)
            raise ValueError(message)
        alchemy_category = self.store.session.query(AlchemyCategory).get(category.pk)
        if not alchemy_category:
            message = _("``Category`` can not be found by the backend.")
            self.store.logger.error(message)
            raise KeyError(message)
        self.store.session.delete(alchemy_category)
        self.store.session.commit()
        message = _("{!r} successfully deleted.".format(category))
        self.store.logger.debug(message)

    def get(self, pk, deleted=None):
        """
        Return a category based on their pk.

        Args:
            pk (int): PK of the category to be retrieved.

        Returns:
            nark.Category: Category matching given PK.

        Raises:
            KeyError: If no such PK was found.

        Note:
            We need this for now, as the service just provides pks, not names.
        """

        message = _("Received PK: '{}'.".format(pk))
        self.store.logger.debug(message)

        if deleted is None:
            result = self.store.session.query(AlchemyCategory).get(pk)
        else:
            query = self.store.session.query(AlchemyCategory)
            query = query.filter(AlchemyCategory.pk == pk)
            query = query_apply_true_or_not(query, AlchemyCategory.deleted, deleted)
            results = query.all()
            assert(len(results) <= 1)
            result = results[0] if results else None

        if not result:
            message = _("No category with 'pk: {}' was found!".format(pk))
            self.store.logger.error(message)
            raise KeyError(message)
        message = _("Returning {!r}.".format(result))
        self.store.logger.debug(message)
        return result.as_hamster(self.store)

    def get_by_name(self, name, raw=False):
        """
        Return a category based on its name.

        Args:
            name (str): Unique name of the category.
            raw (bool): Whether to return the AlchemyCategory instead.

        Returns:
            nark.Category: Category of given name.

        Raises:
            KeyError: If no category matching the name was found.

        """

        message = _("Received name: '{}', raw={}.".format(name, raw))
        self.store.logger.debug(message)

        try:
            result = self.store.session.query(AlchemyCategory).filter_by(name=name).one()
        except NoResultFound:
            message = _("No category named '{}' was found".format(name))
            self.store.logger.debug(message)
            raise KeyError(message)

        if not raw:
            result = result.as_hamster(self.store)
            self.store.logger.debug(_("Returning: {!r}.").format(result))
        return result

    def get_all(self, *args, include_usage=False, sort_col='name', **kwargs):
        """Get all activities."""
        kwargs['include_usage'] = include_usage
        kwargs['sort_col'] = sort_col
        return self._get_all(*args, **kwargs)

    def get_all_by_usage(self, *args, sort_col='usage', **kwargs):
        assert(not args)
        kwargs['include_usage'] = True
        kwargs['sort_col'] = sort_col
        return self._get_all(*args, **kwargs)

    # DRY: This fcn. very much similar between activity/category/tag.
    def _get_all(
        self,
        include_usage=True,
        count_results=False,
        # FIXME/2018-06-20: (lb): Implement since/until.
        since=None,
        until=None,
        # FIXME/2018-06-09: (lb): Implement deleted/hidden. [i.e., in UI]
        deleted=False,
        hidden=False,
        # FIXME/2018-06-20: (lb): Do what with key now?
        key=None,
        search_term='',
        activity=False,
        sort_col='',
        sort_order='',
        # kwargs: limit, offset
        **kwargs
    ):
        """
        Get all Categories, possibly filtered by related Activity, and possible sorted.

        Returns:
            list: List of all Categories present in the database, ordered
            by lower(name) or however caller asked that they be ordered.
        """

        def _get_all_categories():
            message = _('usage: {} / term: {} / act.: {} / col: {} / order: {}'.format(
                include_usage, search_term, activity, sort_col, sort_order,
            ))
            self.store.logger.debug(message)

            query, agg_cols = _get_all_start_query()

            query = _get_all_filter_by_activity(query)

            query = _get_all_filter_by_search_term(query)

            # FIXME/LATER/2018-05-29: (lb): Filter by tags used around this time.
            #   E.g., if it's 4 PM, only suggest tags used on same day at same time...
            #   something like that. I.e., tags you use during weekday at work should
            #   be suggested. For now, filter by category can give similar effect,
            #   depending on how one uses categories.

            query = _get_all_order_by(query, *agg_cols)

            query = _get_all_group_by(query, agg_cols)

            query = query_apply_limit_offset(query, **kwargs)

            query = _get_all_with_entities(query, agg_cols)

            self.store.logger.debug(_('Query') + ': {}'.format(str(query)))

            results = query.all() if not count_results else query.count()

            return results

        def _get_all_start_query():
            agg_cols = []
            if (
                not include_usage
                and not activity
                and sort_col not in ['start', 'activity', ]
            ):
                query = self.store.session.query(AlchemyCategory)
            else:
                count_col = func.count(AlchemyCategory.pk).label('uses')
                agg_cols.append(count_col)

                time_col = func.sum(
                    func.julianday(AlchemyFact.end) - func.julianday(AlchemyFact.start)
                ).label('span')
                agg_cols.append(time_col)

                query = self.store.session.query(AlchemyFact, count_col, time_col)
                query = query.join(AlchemyFact.activity)
                query = query.join(AlchemyCategory)

            return query, agg_cols

        def _get_all_filter_by_activity(query):
            if activity is False:
                return query
            if activity:
                if activity.pk:
                    query = query.filter(AlchemyActivity.pk == activity.pk)
                else:
                    query = query.filter(
                        func.lower(AlchemyActivity.name) == func.lower(activity.name)
                    )
            else:
                query = query.filter(AlchemyFact.activity == None)  # noqa: E711
            return query

        def _get_all_filter_by_search_term(query):
            if not search_term:
                return query
            query = query.filter(
                AlchemyCategory.name.ilike('%{}%'.format(search_term))
            )
            return query

        def _get_all_order_by(query, count_col=None, time_col=None):
            direction = desc if sort_order == 'desc' else asc
            if sort_col == 'start':
                assert include_usage
                direction = desc if not sort_order else direction
                query = query.order_by(direction(AlchemyFact.start))
            elif sort_col == 'usage':
                assert include_usage and count_col is not None
                direction = desc if not sort_order else direction
                query = query.order_by(direction(count_col), direction(time_col))
            elif sort_col == 'time':
                assert include_usage and time_col is not None
                direction = desc if not sort_order else direction
                query = query.order_by(direction(time_col), direction(count_col))
            elif sort_col == 'activity':
                query = query.order_by(direction(AlchemyActivity.name))
                query = query.order_by(direction(AlchemyCategory.name))
            elif sort_col == 'category':
                query = query.order_by(direction(AlchemyCategory.name))
                if count_col is not None:
                    query = query.order_by(direction(AlchemyActivity.name))
            else:
                # FIXME/2018-05-29: (lb): Are all these sort_col's for real?
                # Seems like they wouldn't sort like user would be expecting.
                assert sort_col in ('', 'name', 'tag', 'fact')
                query = query.order_by(direction(AlchemyCategory.name))
                if count_col is not None:
                    query = query.order_by(direction(AlchemyActivity.name))
            return query

        def _get_all_group_by(query, agg_cols):
            if not agg_cols:
                return query
            query = query.group_by(AlchemyCategory.pk)
            return query

        def _get_all_with_entities(query, agg_cols):
            if not agg_cols:
                return query
            query = query.with_entities(AlchemyCategory, *agg_cols)
            return query

        # ***

        return _get_all_categories()




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# All rights reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

from sqlalchemy import Column, ForeignKey, Integer, MetaData, Table

# USAGE: You could run this script manually, invoking SQLAlchemy-migrate, e.g.,
#
#           py3 migrations/manage.py upgrade <db_url> migrations
#
#       Or you could-should run it with the hamster CLI, dob.
#
#           dob migrate up

# NOTE: This script will work against the Legacy Hamster DB on upgrade,
#       but on downgrade, it'll barf an sqlite3.IntegrityError.
#
#       So be sure you upgrade your legacy Hamster database:
#
#           dob store upgrade-legacy ~/.local/share/hamster-applet/hamster.db

# NOTE! We cannot use sqlalchemy.Boolean, lest downgrade barfs!
#
# E.g.,
#
#
#       def upgrade(migrate_engine):
#           ...
#           deleted_c = Column('deleted', Boolean)
#
#       def downgrade(migrate_engine):
#           ...
#           activities.c.deleted.drop()
#
#       $ py3 migrations/manage.py test ${db_url} migrations
#       Upgrading...
#       done
#       Downgrading...
#
#       sqlite3.OperationalError: no such column: deleted
#
#       ...
#
#       sqlalchemy.exc.OperationalError: (sqlite3.OperationalError) no such column:
#       deleted [SQL: '\nCREATE TABLE categories (\n\tid INTEGER NOT NULL,
#                      \n\tname TEXT(500), \n\tcolor_code TEXT(50),
#                      \n\tcategory_order INTEGER, \n\tsearch_name TEXT,
#                      \n\tPRIMARY KEY (id), \n\tCHECK (deleted IN (0, 1))\n)\n\n']
#       (Background on this error at: http://sqlalche.me/e/e3q8)


def upgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)

    activities = Table('activities', meta, autoload=True)
    categories = Table('categories', meta, autoload=True)
    tags = Table('tags', meta, autoload=True)
    facts = Table('facts', meta, autoload=True)

    # SKIP: Activity.deleted already exists.
    upgrade_add_column_boolean(categories, 'deleted')
    upgrade_add_column_boolean(tags, 'deleted')
    upgrade_add_column_boolean(facts, 'deleted')

    # Make a column to store the split-from ID, for edited/split facts.
    # MAYBE: index=True ??
    split_from_id = Column(
        'split_from_id', Integer, ForeignKey(facts.c.id), nullable=True,
    )
    split_from_id.create(facts)

    upgrade_add_column_boolean(activities, 'hidden')
    upgrade_add_column_boolean(categories, 'hidden')
    upgrade_add_column_boolean(tags, 'hidden')
    # SKIP: Facts.hidden probably does not compute.
    #       (Hidden is for auto-complete/MRU lists.)


def downgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)

    activities = Table('activities', meta, autoload=True)
    categories = Table('categories', meta, autoload=True)
    tags = Table('tags', meta, autoload=True)
    facts = Table('facts', meta, autoload=True)

    facts.c.split_from_id.drop()

    facts.c.deleted.drop()
    tags.c.deleted.drop()
    categories.c.deleted.drop()

    # SKIP: facts.c.hidden.drop()
    tags.c.hidden.drop()
    categories.c.hidden.drop()
    activities.c.hidden.drop()


def upgrade_add_column_boolean(table, column):
    # NOTE: (lb): SQLite3 is weird. It preserves the whitespace of
    #       your CREATE TABLE, I think, because after adding the two
    #       columns, deleted and hidden, they're just appended to the
    #       last column's line in the .schema, e.g.,:
    #           sqlite> .schema categories
    #
    #           CREATE TABLE categories (
    #           	id INTEGER NOT NULL,
    #           	name VARCHAR(254), deleted INTEGER, hidden INTEGER,
    #           	PRIMARY KEY (id),
    #           	UNIQUE (name)
    #           );
    #
    # HUH: (lb): Furthermore, SQLAlchemy-migrate preserves whitespace on
    #      upgrade, but now on downgrade.
    #
    #      E.g., support you define you table without newlines, e.g.,:
    #
    #           CREATE TABLE categories (
    #               id INTEGER NOT NULL,
    #               name VARCHAR(254),
    #               PRIMARY KEY (id),
    #               UNIQUE (name)
    #           );
    #
    #      And then you run this script to upgrade. You'll see:
    #
    #           CREATE TABLE categories (
    #               id INTEGER NOT NULL,
    #               name VARCHAR(254),
    #               deleted INTEGER,
    #               hidden INTEGER,
    #               PRIMARY KEY (id),
    #               UNIQUE (name)
    #           );
    #
    #      Then you run this script to downgrade. You'll now see newlines!
    #
    #           CREATE TABLE categories (
    #           	id INTEGER NOT NULL,
    #           	name VARCHAR(254),
    #           	PRIMARY KEY (id),
    #           	UNIQUE (name)
    #           );
    #
    #      Not that any of this matters. It's just interesting.

    # NOTE: We cannot set ``nullable=False`` because existing rows
    #       will violate the constraint. But we can do it soon enough.
    deleted_c = Column(column, Integer, default=False)

    deleted_c.create(table)

    # Get rid of the DEFAULT=FALSE (used to set the cell values)
    # and set NOT NULL.
    deleted_c.alter(nullable=False)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

import os
from datetime import datetime

from .config import decorate_config
from .helpers import logging as logging_helpers
from .helpers.app_dirs import NarkAppDirs
from .managers.activity import BaseActivityManager
from .managers.category import BaseCategoryManager
from .managers.fact import BaseFactManager
from .managers.tag import BaseTagManager

__all__ = ('BaseStore', )


class BaseStore(object):
    """
    A controller store defines the interface to interact with stored entities,
    regardless of the backend being used.
    """

    def __init__(self, config):
        self.config = decorate_config(config)
        self.init_config()
        self.init_logger()
        self._now = None
        self.add_pytest_managers()

    def add_pytest_managers(self):
        if not os.environ.get('PYTEST_CURRENT_TEST', None):
            return
        # The following intermediate classes are solely used for testing!
        self.categories = BaseCategoryManager(self)
        self.activities = BaseActivityManager(self)
        self.tags = BaseTagManager(self)
        localize = self.config['time.tz_aware']
        self.facts = BaseFactManager(self, localize=localize)

    def standup(self):
        """
        Any backend specific setup code that needs to be executed before
        the data store can be used (including creating the data store).
        """
        raise NotImplementedError

    def cleanup(self):
        """
        Any backend specific teardown code that needs to be executed before
        we shut down gracefully.
        """
        raise NotImplementedError

    def init_config(self):
        self.config.setdefault('db.orm', 'sqlalchemy')
        self.config.setdefault('db.engine', 'sqlite')
        self.config.setdefault('db.path', self.default_db_path)
        self.config.setdefault('db.host', '')
        self.config.setdefault('db.port', '')
        self.config.setdefault('db.name', '')
        self.config.setdefault('db.user', '')
        self.config.setdefault('db.password', '')
        self.config.setdefault('time.allow_momentaneous', False)
        self.config.setdefault('time.day_start', '')
        self.config.setdefault('time.fact_min_delta', '0')
        self.config.setdefault('dev.lib_log_level', 'WARNING')
        self.config.setdefault('dev.sql_log_level', 'WARNING')
        self.config.setdefault('time.tz_aware', False)
        self.config.setdefault('time.default_tzinfo', '')

    @property
    def default_db_path(self):
        if NarkAppDirs.APP_DIRS is None:
            return ''
        db_path = os.path.join(
            NarkAppDirs.APP_DIRS.user_data_dir,
            # (lb): Whatever client is using the nark library
            # will generally setup db_path specially; this is
            # just a default filename for completeness.
            'dob.sqlite',
        )
        return db_path

    def init_logger(self):
        sql_log_level = self.config['dev.sql_log_level']
        self.logger = logging_helpers.set_logger_level(
            'nark.store', sql_log_level,
        )

    @property
    def now(self):
        # Use the same 'now' for all items that need it. 'Now' is considered
        # the run of the whole command, and not different points within it.
        # (lb): It probably doesn't matter either way what we do, but I'd
        # like all facts that use now to reflect the same moment in time,
        # rather than being microseconds apart from one another.
        # (lb): Also, we use @property to convey to the caller that this
        # is not a function; i.e., the value is static, not re-calculated.
        if self._now is None:
            self._now = self.now_tz_aware()
        return self._now

    def now_tz_aware(self):
        if self.config['time.tz_aware']:
            # FIXME/2018-05-23: (lb): Tests use utcnow(). Should they honor tz_aware?
            #   (Though if Freezegun being used, now() == utcnow().)
            # Clear microseconds to avoid six digits of noise, e.g., 12:34:56.789012.
            # (lb): I added seconds to hamster (was historically demarcated by minutes),
            # because I think seconds could be useful to a developer. But not no more.
            return datetime.utcnow().replace(microsecond=0)
        else:
            return datetime.now().replace(microsecond=0)




# This file exists within 'nark':
#
#   https://github.com/hotoffthehamster/nark
#
# Copyright © 2018-2020 Landon Bouma
# Copyright © 2015-2016 Eric Goller
# All  rights  reserved.
#
# 'nark' is free software: you can redistribute it and/or modify it under the terms
# of the GNU General Public License  as  published by the Free Software Foundation,
# either version 3  of the License,  or  (at your option)  any   later    version.
#
# 'nark' is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY  or  FITNESS FOR A PARTICULAR
# PURPOSE.  See  the  GNU General Public License  for  more details.
#
# You can find the GNU General Public License reprinted in the file titled 'LICENSE',
# or visit <http://www.gnu.org/licenses/>.

"""Factories which generate ready-to-use, randomized nark class object instances."""

import datetime

# BEWARE: This module imports packages that are installed for developers only,
# via requirements/test.pip, and are not specified as package dependencies. So do
# not load from this module from any nark-proper code! Normal users are not to be
# expected to have these libraries available. Only load this module from other tests.

import factory
import fauxfactory

from nark.items.activity import Activity
from nark.items.category import Category
from nark.items.fact import Fact
from nark.items.tag import Tag


# 2020-01-30: (lb): DRYing time: Here's a little history, because I made some
# assumptions when I merged the two, not-so-DRY, factories.py from nark and dob.
# - There were two sources of divergence:
#   - hamster-lib's factories.py used factory.Faker(), instead of faker(),
#     for Fact.start.
#   - hamster-lib's factories.py used factory.LazyAttribute and fauxfactory
#     to generate Category.name, instead of factory.Faker.
# - I think hamster-lib had the most up to date code -- the original behavior
#   was committed 2015-12-17 to hamster-lib, and later copied to hamster-cli.
#   Then, changes were made to the factories.py module in hamster-lib, but not
#   to the copy-pasta version in hamster-cli. From hamster-lib's log:
#       2016-06-24 hamster-lib elbenfreund c84f189a
#       This removes the naive and error prone usage of stand alone ``faker``
#       calls in our factories for the proper ``factory.Faker`` ones. Those
#       basicly wrap our calls in ``LazyAttributes``. Where we generate random
#       data from non-faker sources, e.g. ``fauxfactory`` we make sure to do the
#       wrapping ourselfs.
#   I don't quite understand what the issue was, but sounds like the
#   solution is to not use faker() (from `import faker`), but to use
#   factory.Faker() (and also factory.LazyAttribute with fauxfactory).

class CategoryFactory(factory.Factory):
    """Test fixture factory returns new ``Category`` with random attribute values."""

    pk = None
    # 2020-01-30: (lb): DRYing time: I consolidated factories.py from nark and dob
    # and saw this code from dob (which came from hamster-cli, which was copied from
    # hamster-lib; but the hamster-cli copy is missing a bugfix that hamster-lib's
    # copy received; here's the original code):
    #
    #   name = factory.Faker('word')
    #
    # Here's a comment from elbenfreund, from tests/hamster_lib/factories.py,
    # regarding the newer use of LazyAttribute:
    #
    #   # Although we do not need to reference to the object beeing created and
    #   # ``LazyFunction`` seems sufficient it is not as we could not pass on the
    #   # string encoding. ``LazyAttribute`` allows us to specify a lambda that
    #   # circumvents this problem.
    name = factory.LazyAttribute(lambda x: fauxfactory.gen_string('utf8'))

    class Meta:
        model = Category


class ActivityFactory(factory.Factory):
    """Test fixture factory returns new ``Activity`` with random attribute values."""

    pk = None
    name = factory.Faker('word')
    category = factory.SubFactory(CategoryFactory)
    deleted = False

    class Meta:
        model = Activity


class TagFactory(factory.Factory):
    """Test fixture factory returns new ``Tag`` with random attribute values."""

    pk = None
    name = factory.Faker('word')

    class Meta:
        model = Tag


class FactFactory(factory.Factory):
    """Test fixture factory returns new ``Fact`` with random attribute values.

    Each new instance has a three-hour duration (i.e., the start time is
    randomized, and the end time is calculated three hours after that).
    """

    pk = None
    activity = factory.SubFactory(ActivityFactory)

    # 2020-01-30: (lb): DRYing time: I consolidated factories.py from nark and dob,
    # and here's the second discrepancy I found (where I think hamster-cli was
    # overlooked by a bugfix applied to hamster-lib). The original code was committed
    # way back when and used faker:
    #   # 2015-12-17 hamster-lib tests/factories.py 52976ac0
    #   import faker
    #   start = faker.Faker().date_time()
    #   end = start + datetime.timedelta(hours=3)
    # and then hamster-lib (but not the -cli) was updated 2016-06-24 (c84f189a) thusly:
    start = factory.Faker('date_time')
    end = factory.LazyAttribute(lambda o: o.start + datetime.timedelta(hours=3))
    description = factory.Faker('paragraph')

    class Meta:
        model = Fact

    @factory.post_generation
    def tags(self, create, extracted, **kwargs):
        """Add new random tags after instance creation."""
        self.tags = set([TagFactory() for i in range(1)])

